// Initial Developer: 2002 Takeshi Morimoto <morimoto@takopen.cs.uec.ac.jp> 
// Author: Zijian Ren
// Version :1.30
// Date: Feb 24, 2004
// Comments: Add police route: not considering road blockades for polices can clear them
//		Add protected methods: checkTrafficJam() and move_N(). Modify move()
//		Add members: actionTaken, prevLocation
//		Add method moveToBuiling() and mlpos()

package yab.agent;

import java.net.*;
import java.util.*;
import yab.agent.*;
import yab.agent.object.*;

public abstract class HumanoidAgent extends Agent {
	protected String actionTaken = "";
	protected MotionlessObject prevLocation;
	protected Route prevRoute;
	
    protected HumanoidAgent(int agentType, InetAddress address, int port) {
        super(agentType, address, port);
    }

    private Humanoid self() { return (Humanoid) world.self; }

    protected int hearingLimit() { return HEARING_LIMIT_OF_HUMANOID; }

    protected MotionlessObject mlpos() { return self().motionlessPosition(); }
    
    protected boolean checkTrafficJam() {
        if (actionTaken.equals("move") && time() > Constants.TIME_STARTING_ACTION+1) {
		//Stuck in this position
		if (mlpos() == prevLocation && mlpos() instanceof Road) {
			System.out.println("Time "+time()+" "+self()+" is in traffic jam at "+mlpos());
			return true;
		}
	}
    
	return false;	

    }
    
    protected void moveToBuilding() throws ActionCommandException {
    	if (!(mlpos() instanceof Road))
		return;
	
	HashSet nearbyBuildings = new HashSet();
			
	for(Iterator it = world.buildings.iterator(); it.hasNext(); ) {
		Building b = (Building)it.next();
		if ( b.time()==time() && !b.isBurning() )
			nearbyBuildings.add(b);
	}
	
	if (!nearbyBuildings.isEmpty())
		move(nearbyBuildings);
	else {
		System.out.println("NO NEARBY BUILDING AVAILABLE!");	
		move(world.refuges);	
	}	
    }
    
    protected void move(MotionlessObject destination)
        throws ActionCommandException {
        move(Collections.singleton(destination));
    }
    protected void move(Collection destinations) throws ActionCommandException{
        move(getRoute(destinations));
    }
    protected void move(Route routePlan) throws ActionCommandException {
    //the route is current route but will be previous route at the next cycle
	actionTaken = "move";
    	prevRoute = routePlan;
	prevLocation = self().motionlessPosition();
	
        move(routePlan.toIDs());
    }
    protected void move(int[] routePlan) throws ActionCommandException {
        socket().akMove(self().id, routePlan);
        throw new ActionCommandException();
    }

    protected void move_N(Collection destinations) throws ActionCommandException{
    	move(destinations);
    }
    
    /** This method returns the minimum cost route reaching one of
     *  <code>destinations</code> in consideration for a blockade
     *  located at the midpoint of a road where the controlled object
     *  is.
     *  This method routes, estimating cost of a route based upon its
     *  passability:
     *  <ul>
     *      <li>if the agent has seen a road and the road is passable,
     *      cost for reaching the road is low</li>
     *      <li>if the agent has seen a road and the road is not
     *      passable, the cost is high</li>
     *      <li>if the agent has not seen a road but the road may be
     *      passable, the cost is the middle</li>
     *  </ul>
     * @see Router.CostFunction#cost(MotionlessObject, MotionlessObject) cost
     */
    protected Route getRoute(Collection destinations) {
        MotionlessObject pos = self().motionlessPosition();
        if (!(pos instanceof Road)  ||  ((Road) pos).passableLines() >= 1)
            return Router.get(pos, destinations, RELIABILITY_COST_FUNCTION);
        Route route = Router.get(outsidePosition(), destinations,
                                 RELIABILITY_COST_FUNCTION);
        modifyOriginToCurrentPosition(route);
        return route;
    }
    private MotionlessObject outsidePosition() {
        Road rd = (Road) self().motionlessPosition();
        return (self().positionExtra() < rd.length() / 2)
            ? rd.head() : rd.tail();
    }
    private void modifyOriginToCurrentPosition(Route route) {
        MotionlessObject pos = self().motionlessPosition();
        for (Route rt = route;  true;  rt = rt.previous()) {
            if (rt.end == pos) {
                rt.setPrevious(null);
                return;
            }
            if (rt.previous() == null) {
                rt.setPrevious(new Route(pos));
                return;
            }
        }
    }
    private static final Router.CostFunction RELIABILITY_COST_FUNCTION
        = new Router.CostFunction() {
                public float cost(MotionlessObject from, MotionlessObject to) {
                    final int WEIGHT = Integer.MAX_VALUE;
                    float c = 1;
                    if (!(to instanceof Road))
                        return c;
                    Road rd = (Road) to;
                    PointObject po = (PointObject) from;
                    if (rd.passableLinesFrom(po) == 0)
                        return c * WEIGHT * WEIGHT;
                    return rd.hasBeenSeen() ? c : c * WEIGHT; }};
}
